@@ -2,13 +2,12 @@
 This module implements generic, reusable calibration models that can be subclassed to
 implement custom calibration models.
 """
-from collections import namedtuple
 import numpy
 import scipy
 import typing
 
 from .. import core
-from .. import utils 
+from .. import utils
 
 try:
     import theano
@@ -22,13 +21,15 @@ except ModuleNotFoundError:
 
 class BaseModelT(core.CalibrationModel):
     def loglikelihood(self, *, y, x, replicate_id: str=None, dependent_key: str=None, theta=None):
-        """Loglikelihood of observation (dependent variable) given the independent variable
+        """ Loglikelihood of observation (dependent variable) given the independent variable.
+
+        If both x and y are vectors, they must have the same length and the likelihood will be evaluated elementwise.
 
         Parameters
         ----------
-        y : array-like
+        y : scalar or array-like
             observed measurements (dependent variable)
-        x : array-like or TensorVariable
+        x : scalar, array-like or TensorVariable
             assumed independent variable
         replicate_id : optional, str
             unique identifier for replicate (necessary for pymc3 likelihood)
@@ -46,6 +47,16 @@ class BaseModelT(core.CalibrationModel):
             if self.theta_fitted is None:
                 raise Exception('No parameter vector was provided and the model is not fitted with data yet.')
             theta = self.theta_fitted
+
+        if not isinstance(x, (list, numpy.ndarray, float, int)) and not utils.istensor(x):
+            raise ValueError(
+                f'Input x must be a scalar, TensorVariable or an array-like object, but not {type(x)}'
+            )
+        if not isinstance(y, (list, numpy.ndarray, float, int)) and not utils.istensor(x):
+            raise ValueError(
+                f'Input y must be a scalar or an array-like object, but not {type(y)}'
+            )
+
         mu, scale, df = self.predict_dependent(x, theta=theta)
         if utils.istensor(x) or utils.istensor(theta):
             if pm.Model.get_context(error_if_none=False) is not None:
@@ -68,98 +79,8 @@ class BaseModelT(core.CalibrationModel):
                     nu=df,
                 ).logp(y).sum()
             return L
-        elif isinstance(x, (list, numpy.ndarray)):
-            # using t-distributed noise in the non-transformed space
-            loglikelihoods = scipy.stats.t.logpdf(x=y, loc=mu, scale=scale, df=df)
-            return numpy.sum(loglikelihoods)
         else:
-            raise Exception('Input x must either be a TensorVariable or an array-like object.')
-
-    def infer_independent(
-        self, y:typing.Union[int,float,numpy.ndarray], *, 
-        lower:float, upper:float, steps:int=300, 
-        hdi_prob:float=1
-    ) -> core.NumericPosterior:
-        """Infer the posterior distribution of the independent variable given the observations of the dependent variable.
-        The calculation is done numerically by integrating the likelihood in a certain interval [upper,lower]. 
-        This is identical to the posterior with a Uniform (lower,upper) prior. If precentiles are provided, the interval of
-        the PDF will be shortened.
-
-        Parameters
-        ----------
-        y : int, float, array
-            one or more observations at the same x
-        lower : float
-            lower limit for uniform distribution of prior
-        upper : float
-            upper limit for uniform distribution of prior
-        steps : int
-            steps between lower and upper or steps between the percentiles (default 300)
-        hdi_prob : float
-            if 1 (default), the complete interval [upper,lower] will be returned, 
-            else pdf will be trimmed to the according probability interval; 
-            float must be in the interval (0,1]
-                                
-        Returns
-        -------
-        posterior : NumericPosterior
-            the result of the numeric posterior calculation
-        """  
-        y = numpy.atleast_1d(y)
-
-        def likelihood(x, y):
-            loc, scale, df = self.predict_dependent(x)
-            # get log-probs for all observations
-            logpdfs = [
-                scipy.stats.t.logpdf(y_, loc=loc, scale=scale, df=df)
-                for y_ in y
-            ]
-            # sum them and exp them (numerically better than numpy.prod of pdfs)
-            return numpy.exp(numpy.sum(logpdfs, axis=0))
-        
-        # high resolution x-coordinates for integration
-
-        likelihood_integral, _ = scipy.integrate.quad(
-            func=likelihood,
-            # by restricting the integral into the interval [a,b], the resulting PDF is
-            # identical to the posterior with a Uniform(a, b) prior.
-            # 1. prior probability is constant in [a,b]
-            # 2. prior probability is 0 outside of [a,b]
-            # > numerical integral is only computed in [a,b], but because of 1. and 2., it's
-            #   identical to the integral over [-∞,+∞]
-                a=lower, b=upper,
-                args=(y,)
-            )
-
-        x_integrate = numpy.linspace(lower, upper, 100_000)
-        area_by_x = scipy.integrate.cumtrapz(likelihood(x_integrate, y), x_integrate, initial=0)
-        prob_by_x = area_by_x / area_by_x[-1]
-        
-        if hdi_prob != 1:
-            if not (0 < hdi_prob <= 1):
-                raise ValueError(f'Unexpected `hdi_prob` value of {hdi_prob}. Expected float in interval (0, 1].')
-
-            i_lower = numpy.argmax(prob_by_x > (1 - hdi_prob)/2)
-            i_upper = numpy.argmax(prob_by_x > (1 + hdi_prob)/2)
-            x_dense = numpy.linspace(
-                x_integrate[i_lower],
-                x_integrate[i_upper-1],
-                steps
-            )       
-        else:
-            x_dense = numpy.linspace(lower, upper, steps)
-
-        pdf = likelihood(x_dense, y) / likelihood_integral
-        # find indices for median
-        i_501 = numpy.argmax(prob_by_x > 0.5)
-        median = numpy.mean([
-            x_integrate[i_501],
-            x_integrate[i_501-1]
-        ])
-        lower_x = numpy.min(x_dense)
-        upper_x = numpy.max(x_dense)
-        data = core.NumericPosterior(x_dense, pdf, median, hdi_prob, lower_x, upper_x)
-        return data
+            return numpy.sum(scipy.stats.t.logpdf(x=y, loc=mu, scale=scale, df=df))
 
 
 class BasePolynomialModelT(BaseModelT):
