@@ -4,64 +4,211 @@ This module contains type definitions that generalize across all applications.
 Also, it implements a variety of modeling functions such as polynomials,
 or (asymmetric) logistic functions and their corresponding inverse functions.
 """
-import abc
+from collections import defaultdict
 import datetime
 import inspect
 import json
 import logging
 import numpy
 import os
-import scipy.optimize
+import scipy
 import typing
 
 from . import utils
 
 
-__version__ = '5.0.3'
+__version__ = '6.0.0'
 _log = logging.getLogger('calibr8')
 
 
-class NumericPosterior(typing.NamedTuple):
-    """ The result of a numeric infer_independent operation.
+class NumericPosterior:
+    """ The result of a numeric infer_independent operation. """
+    def __init__(
+        self,
+        median: float,
+        eti_x: numpy.ndarray,
+        eti_pdf: numpy.ndarray,
+        eti_prob: float,
+        hdi_x: numpy.ndarray,
+        hdi_pdf: numpy.ndarray,
+        hdi_prob: float,
+    ) -> None:
+        """ The result of a numeric infer_independent operation.
 
-    Attributes
-    ----------
-    x_dense : array
-        values of the independent variable in the percentiles or in [lower, upper]
-    pdf : array
-        values of the posterior pdf at positions [x]
-    median : float
-        x-value of the posterior median
-    hdi_prob: float
-        highest density interval probability (0,1]
-    lower : float
-        x-value at the lower bound of the hdi
-    upper : float
-        x-value at the upper bound of the hdi
-    """
-    x_dense: numpy.ndarray
-    pdf: numpy.ndarray
-    median: float
-    hdi_prob: float
-    lower: float
-    upper: float
+        Parameters
+        ----------
+        median : float
+            x-value of the posterior median
+        eti_x : array
+            Values of the independent variable in [eti_lower, eti_upper]
+        eti_pdf : array
+            Values of the posterior pdf at positions [eti_x]
+        eti_prob : float
+            Probability mass in the ETI
+        hdi_x : array
+            Values of the independent variable in [hdi_lower, hdi_upper]
+        hdi_pdf : array
+            Values of the posterior pdf at positions [hdi_x]
+        hdi_prob : float
+            Probability mass in the HDI
+        """
+        self.median = median
+        self.eti_x = eti_x
+        self.eti_pdf = eti_pdf
+        self.eti_prob = eti_prob
+        self.hdi_x = hdi_x
+        self.hdi_pdf = hdi_pdf
+        self.hdi_prob = hdi_prob
 
     def __repr__(self) -> str:
-        result = str(type(self))
-        for name, value in zip(self._fields, self):
-            result += f"\n    .{name}:\t{type(value).__name__}"
+        result = (
+            str(type(self))
+            + f"\n    ETI ({round(self.eti_prob, 3) * 100:.1f} %): [{round(self.eti_lower, 4)}, {round(self.eti_upper, 4)}] Δ={round(self.eti_width, 4)}"
+            + f"\n    HDI ({round(self.hdi_prob, 3) * 100:.1f} %): [{round(self.hdi_lower, 4)}, {round(self.hdi_upper, 4)}] Δ={round(self.hdi_width, 4)}"
+        )
         return result
 
+    @property
+    def eti_lower(self) -> float:
+        """ Lower bound of the ETI. This is the first value in `eti_x`. """
+        return self.eti_x[0]
+
+    @property
+    def eti_upper(self) -> float:
+        """ Upper bound of the ETI. This is the last value in `eti_x`. """
+        return self.eti_x[-1]
+
+    @property
+    def eti_width(self) -> float:
+        """ Width of the ETI. """
+        return self.eti_upper - self.eti_lower
+
     @property
     def hdi_lower(self) -> float:
-        """ Lower bound of the HDI. This is the first value in `x_dense`. """
-        return self.x_dense[0]
+        """ Lower bound of the HDI. This is the first value in `hdi_x`. """
+        return self.hdi_x[0]
 
     @property
     def hdi_upper(self) -> float:
-        """ Upper bound of the HDI. This is the last value in `x_dense`. """
-        return self.x_dense[-1]
+        """ Upper bound of the HDI. This is the last value in `hdi_x`. """
+        return self.hdi_x[-1]
 
+    @property
+    def hdi_width(self) -> float:
+        """ Width of the HDI. """
+        return self.hdi_upper - self.hdi_lower
+
+
+def _interval_prob(x_cdf: numpy.ndarray, cdf: numpy.ndarray, a: float, b: float):
+    """Calculates the probability in the interval [a, b]."""
+    ia = numpy.argmin(numpy.abs(x_cdf - a))
+    ib = numpy.argmin(numpy.abs(x_cdf - b))
+    return (cdf[ib] - cdf[ia])
+
+
+def _get_eti(
+    x_cdf: numpy.ndarray,
+    cdf: numpy.ndarray,
+    ci_prob: float
+) -> typing.Tuple[float, float]:
+    """ Find the equal tailed interval (ETI) corresponding to a certain credible interval probability level.
+
+    Parameters
+    ----------
+    x_cdf : numpy.ndarray
+        Coordinates where the cumulative density function was evaluated
+    cdf : numpy.ndarray
+        Values of the cumulative density function at `x_cdf`
+    ci_prob : float
+        Desired probability level
+
+    Returns
+    -------
+    eti_lower : float
+        Lower bound of the ETI
+    eti_upper : float
+        Upper bound of the ETI
+    """
+    i_lower = numpy.argmin(numpy.abs(cdf - (1 - ci_prob) / 2))
+    i_upper = numpy.argmin(numpy.abs(cdf - (1 + ci_prob) / 2))
+    eti_lower = x_cdf[i_lower]
+    eti_upper = x_cdf[i_upper]
+    return eti_lower, eti_upper
+
+
+def _get_hdi(
+    x_cdf: numpy.ndarray,
+    cdf: numpy.ndarray,
+    ci_prob: float,
+    guess_lower: float,
+    guess_upper: float,
+    *,
+    history: typing.Optional[typing.DefaultDict[str, typing.List]]=None
+) -> typing.Tuple[float]:
+    """ Find the highest density interval (HDI) corresponding to a certain credible interval probability level.
+
+    Parameters
+    ----------
+    x_cdf : numpy.ndarray
+        Coordinates where the cumulative density function was evaluated
+    cdf : numpy.ndarray
+        Values of the cumulative density function at `x_cdf`
+    ci_prob : float
+        Desired probability level
+    guess_lower : float
+        Initial guess for the lower bound of the HDI
+    guess_upper : float
+        Initial guess for the upper bound of the HDI
+    history : defaultdict of list, optional
+        A defaultdict(list) may be passed to capture intermediate parameter and loss values
+        during the optimization. Helps to understand, diagnose and test.
+
+    Returns
+    -------
+    hdi_lower : float
+        Lower bound of the HDI
+    hdi_upper : float
+        Upper bound of the HDI
+    """
+    def hdi_objective(x):
+        a, d = x
+        b = a + d
+
+        prob = _interval_prob(x_cdf, cdf, a, b)
+        delta_prob = numpy.abs(prob - ci_prob)
+
+        if prob < ci_prob:
+            # do not allow shrinking below the desired level
+            L_prob = numpy.inf
+            L_delta = 0
+        else:
+            # above the desired level penalize the interval width
+            L_prob = 0
+            L_delta = d
+
+        L = L_prob + L_delta
+
+        if history is not None:
+            history["prob"].append(prob)
+            history["delta_prob"].append(delta_prob)
+            history["a"].append(a)
+            history["b"].append(b)
+            history["d"].append(d)
+            history["L_prob"].append(L_prob)
+            history["L_delta"].append(L_delta)
+            history["L"].append(L)
+        return L
+
+    fit = scipy.optimize.fmin(
+        hdi_objective,
+        # parametrize as b=a+d
+        x0=[guess_lower, guess_upper - guess_lower],
+        xtol=numpy.ptp(x_cdf) / len(x_cdf),
+        disp=False
+    )
+    hdi_lower, hdi_width = fit
+    hdi_upper = hdi_lower + hdi_width
+    return hdi_lower, hdi_upper
 
 class CalibrationModel:
     """A parent class providing the general structure of a calibration model."""
@@ -162,13 +309,12 @@ class CalibrationModel:
     def infer_independent(
         self, y:typing.Union[int,float,numpy.ndarray], *, 
         lower:float, upper:float, steps:int=300, 
-        hdi_prob:float=1
+        ci_prob:float=1
     ) -> NumericPosterior:
         """Infer the posterior distribution of the independent variable given the observations of the dependent variable.
-
-        The calculation is done numerically by integrating the likelihood in a certain interval [upper, lower].
-        This is identical to the posterior with a Uniform (lower, upper) prior. If a hdi_prob < 1 is specified,
-        the interval of the PDF will be shortened.
+        The calculation is done numerically by integrating the likelihood in a certain interval [upper,lower]. 
+        This is identical to the posterior with a Uniform (lower,upper) prior. If precentiles are provided, the interval of
+        the PDF will be shortened.
 
         Parameters
         ----------
@@ -180,31 +326,84 @@ class CalibrationModel:
             upper limit for uniform distribution of prior
         steps : int
             steps between lower and upper or steps between the percentiles (default 300)
-        hdi_prob : float
-            if 1 (default), the complete interval [upper,lower] will be returned, 
-            else pdf will be trimmed to the according probability interval; 
+        ci_prob : float
+            Probability level for ETI and HDI credible intervals.
+            If 1 (default), the complete interval [upper,lower] will be returned, 
+            else the PDFs will be trimmed to the according probability interval; 
             float must be in the interval (0,1]
-                                
+
         Returns
         -------
         posterior : NumericPosterior
             the result of the numeric posterior calculation
         """  
-        raise NotImplementedError('The infer_independent function should be implemented by the inheriting class.')
+        y = numpy.atleast_1d(y)
+
+        likelihood_integral, _ = scipy.integrate.quad(
+            func=lambda x: self.likelihood(x=x, y=y),
+            # by restricting the integral into the interval [a,b], the resulting PDF is
+            # identical to the posterior with a Uniform(a, b) prior.
+            # 1. prior probability is constant in [a,b]
+            # 2. prior probability is 0 outside of [a,b]
+            # > numerical integral is only computed in [a,b], but because of 1. and 2., it's
+            #   identical to the integral over [-∞,+∞]
+            a=lower, b=upper,
+        )
+
+        # high resolution x-coordinates for integration
+        # the first integration is just to find the peak
+        x_integrate = numpy.linspace(lower, upper, 10_000)
+        area = scipy.integrate.cumtrapz(self.likelihood(x=x_integrate, y=y, scan_x=True), x_integrate, initial=0)
+        cdf = area / area[-1]
+
+        # now we find a high-resolution CDF for (1-shrink) of the probability mass
+        shrink = 0.00001
+        xfrom, xto = _get_eti(x_integrate, cdf, 1 - shrink)
+        x_integrate = numpy.linspace(xfrom, xto, 100_000)
+        area = scipy.integrate.cumtrapz(self.likelihood(x=x_integrate, y=y, scan_x=True), x_integrate, initial=0)
+        cdf = (area / area[-1]) * (1 - shrink) + shrink / 2
+
+        # TODO: create a smart x-vector from the CDF with varying stepsize
+
+        if ci_prob != 1:
+            if not (0 < ci_prob <= 1):
+                raise ValueError(f'Unexpected `ci_prob` value of {ci_prob}. Expected float in interval (0, 1].')
+
+            # determine the interval bounds from the high-resolution CDF
+            eti_lower, eti_upper = _get_eti(x_integrate, cdf, ci_prob)
+            hdi_lower, hdi_upper = _get_hdi(x_integrate, cdf, ci_prob, eti_lower, eti_upper, history=None)
+
+            eti_x = numpy.linspace(eti_lower, eti_upper, steps)
+            hdi_x = numpy.linspace(hdi_lower, hdi_upper, steps)
+            eti_pdf = self.likelihood(x=eti_x, y=y, scan_x=True) / likelihood_integral
+            hdi_pdf = self.likelihood(x=hdi_x, y=y, scan_x=True) / likelihood_integral
+            eti_prob = _interval_prob(x_integrate, cdf, eti_lower, eti_upper)
+            hdi_prob = _interval_prob(x_integrate, cdf, hdi_lower, hdi_upper)
+        else:
+            x = numpy.linspace(lower, upper, steps)
+            eti_x = hdi_x = x
+            eti_pdf = hdi_pdf = self.likelihood(x=x, y=y, scan_x=True) / likelihood_integral
+            eti_prob = hdi_prob = 1
+
+        median = x_integrate[numpy.argmin(numpy.abs(cdf - 0.5))]
+
+        return NumericPosterior(
+            median,
+            eti_x, eti_pdf, eti_prob,
+            hdi_x, hdi_pdf, hdi_prob,
+        )
 
     def loglikelihood(self, *, y, x, theta=None):
-        """Loglikelihood of observation (dependent variable) given the independent variable
+        """ Loglikelihood of observation (dependent variable) given the independent variable.
+
+        If both x and y are vectors, they must have the same length and the likelihood will be evaluated elementwise.
 
         Parameters
         ----------
-        y : array-like
+        y : scalar or array-like
             observed measurements (dependent variable)
-        x : array-like or TensorVariable
+        x : scalar, array-like or TensorVariable
             assumed independent variable
-        replicate_id : optional, str
-            unique identifier for replicate (necessary for pymc3 likelihood)
-        dependent_key : optional, str
-            key of the dependent variable (necessary for pymc3 likelihood)
         theta : optional, array-like
             model parameters
 
@@ -215,6 +414,34 @@ class CalibrationModel:
         """
         raise NotImplementedError('The loglikelihood function should be implemented by the inheriting class.')
 
+    def likelihood(self, *, y, x, theta=None, scan_x: bool=False):
+        """ Likelihood of observation (dependent variable) given the independent variable.
+
+        Relies on the `loglikelihood` method.
+
+        Parameters
+        ----------
+        y : scalar or array-like
+            observed measurements (dependent variable)
+        x : scalar, array-like or TensorVariable
+            assumed independent variable
+        theta : optional, array-like
+            model parameters
+        scan_x : bool
+            When set to True, the method evaluates `likelihood(xi, y) for all xi in x`
+
+        Returns
+        -------
+        L : float or TensorVariable
+            sum of likelihoods
+        """
+        if scan_x:
+            return numpy.exp([
+                self.loglikelihood(y=y, x=xi, theta=theta)
+                for xi in x
+            ])
+        return numpy.exp(self.loglikelihood(y=y, x=x, theta=theta))
+
     def objective(self, independent, dependent, minimize=True) -> typing.Callable:
         """Creates an objective function for fitting to data.
         
